# Copyright 2019-2020 John Lees, Sam Horsfield

'''Detect presence of sequence elements using graphs/indexes'''

import os, sys

import tempfile
from multiprocessing import Pool
from functools import partial

import unitig_query

from .__init__ import __version__

from .bifrost import rtab_format
from .bifrost import pyseer_format

def get_options():
    import argparse

    description = 'Call unitigs in a population dataset'
    parser = argparse.ArgumentParser(description=description,
                                     prog='unitig-caller')

    modeGroup = parser.add_argument_group('Mode of operation')
    mode = modeGroup.add_mutually_exclusive_group(required=True)
    mode.add_argument('--call',
                        action='store_true',
                        default=False,
                        help='Build a DBG and call colours of unitigs within ')
    mode.add_argument('--query',
                        action='store_true',
                        default=False,
                        help='Query unitig colours in reference genomes/DBG ')
    mode.add_argument('--simple',
                        action='store_true',
                        default=False,
                        help='Use FM-index to make calls ')

    io = parser.add_argument_group('Unitig-caller input/output')
    io.add_argument('--refs',
                    help='Ref file to used to build DBG or use with --simple',
                    default=None)
    io.add_argument('--reads',
                    help='Read file to used to build DBG',
                    default=None)
    io.add_argument('--graph',
                    help='Existing graph in GFA format',
                    default=None)
    io.add_argument('--colours',
                    help='Existing bifrost colours file in .bfg_colors format',
                    default=None)
    io.add_argument('--unitigs',
                    help='Text or fasta file of unitigs to query (--query or --simple)',
                    default=None)
    io.add_argument('--pyseer',
                    action='store_true',
                    help='Output pyseer format',
                    default=False)
    io.add_argument('--rtab',
                    action='store_true',
                    help='Output rtab format',
                    default=False)
    io.add_argument('--out',
                    default='unitig_caller',
                    help='Prefix for output '
                         '[default = \'unitig_caller\']')

    bifrost = parser.add_argument_group('Bifrost options')
    bifrost.add_argument('--kmer',
                        type=int,
                        default=31,
                        help='K-mer size for graph building/querying '
                             '[default = 31]')
    bifrost.add_argument('--write-graph',
                        action='store_true',
                        default=False,
                        help='Output DBG built with unitig-caller')

    simple = parser.add_argument_group('Simple mode options')
    simple.add_argument('--no-save-idx',
                default=False,
                action='store_true',
                help='Do not save FM-indexes for reuse')

    other = parser.add_argument_group('Other')
    other.add_argument('--threads',
                        type=int,
                        default=1,
                        help='Number of threads to use '
                             '[default = 1]')
    other.add_argument('--version', action='version',
                       version='%(prog)s '+__version__)

    return parser.parse_args()


def main():
    options = get_options()

    if options.call or options.query:
        if options.call and options.graph != None and options.colours != None and options.refs == None and options.reads == None:
            # Read input1 (and input2 if specified) as reads.txt or refs.txt. Call `Bifrost build`

            sys.stderr.write("Calling unitigs within input genomes...\n")
            unitig_map, input_colour_pref = unitig_query.call_unitigs_existing(options.graph, options.colours, True, "NA", options.threads)

        elif options.call and (options.refs != None or options.reads != None) and options.graph == None and options.colours == None:

            sys.stderr.write("Building a DBG and calling unitigs within...\n")
            if options.refs != None and options.reads == None:
                unitig_map, input_colour_pref = unitig_query.call_unitigs_build(options.refs, options.kmer, True, "NA", options.threads, True, options.write_graph)
            elif options.refs == None and options.reads != None:
                unitig_map, input_colour_pref = unitig_query.call_unitigs_build(options.reads, options.kmer, True, "NA", options.threads, False, options.write_graph)
            elif options.refs != None and options.reads != None:
                unitig_map, input_colour_pref = unitig_query.call_unitigs_build(options.refs, options.kmer, True, "NA", options.threads, False, options.write_graph, options.reads)

        elif options.query and options.graph != None and options.colours != None and options.unitigs != None and options.refs == None and options.reads == None:

            sys.stderr.write("Querying unitigs within existing DBG...\n")
            unitig_map, input_colour_pref = unitig_query.call_unitigs_existing(options.graph, options.colours, False, options.unitigs, options.threads)

        elif options.query and (options.refs != None or options.reads != None) and options.unitigs != None and options.graph == None and options.colours == None:
            sys.stderr.write("Building a DBG and querying unitigs within...\n")
            if options.refs != None and options.reads == None:
                unitig_map, input_colour_pref = unitig_query.call_unitigs_build(options.refs, options.kmer, False, options.unitigs, options.threads, True, options.write_graph)
            elif options.refs == None and options.reads != None:
                unitig_map, input_colour_pref = unitig_query.call_unitigs_build(options.reads, options.kmer, False, options.unitigs, options.threads, False, options.write_graph)
            elif options.refs != None and options.reads != None:
                unitig_map, input_colour_pref = unitig_query.call_unitigs_build(options.refs, options.kmer, False, options.unitigs, options.threads, False, options.write_graph, options.reads)

        else:
            print("Error: incorrect number of input files specified. Please only specify the below combinations:\n"
                "For --call:\n"
                "   - Bifrost GFA and Bifrost colours file (--graph & --colours)\n"
                "   - List of reference files (--refs)\n"
                "   - List of read files (--reads)\n"
                "   - A list of reference files and a list of read files (--refs & --reads)\n"
                "For --query:\n"
                "   - One of the above combinations with a text file with header, one file per line/fasta file of query unitigs (--unitigs)\n")
            sys.exit(1)

        if options.pyseer or (not options.pyseer and not options.rtab):
            sys.stderr.write("Generating pyseer file...\n")
            pyseer_format(unitig_map, input_colour_pref, options.out + ".pyseer")

        if options.rtab:
            sys.stderr.write("Generating rtab file...\n")
            rtab_format(unitig_map, input_colour_pref, options.out + ".rtab")

    elif options.simple:
        # Read input into lists, as in 'index' and 'call'

        if options.refs == None or options.unitigs == None:
            sys.stderr.write("Error: Please specify a list of assemblies as --refs and unitigs file as --unitigs\n")
            sys.exit(1)
        else:
            names_in = []
            fasta_in = []
            with open(options.refs, 'r') as strain_file:
                for strain_line in strain_file:
                    strain_line = strain_line.rstrip()
                    names_in.append(os.path.splitext(os.path.basename(strain_line))[0])
                    fasta_in.append(strain_line)

            unitigs = []
            with open(options.unitigs, 'r') as unitig_file:
                unitig_file.readline() # header
                for unitig_line in unitig_file:
                    if unitig_line[0] != '>':
                        unitig_fields = unitig_line.rstrip().split("\t")
                        unitigs.append(unitig_fields[0])

            # call c++ code to map (also writes output file)
            unitig_query.call(fasta_in,
                             names_in,
                             unitigs,
                             options.out + ".pyseer",
                             not options.no_save_idx,
                             options.threads)

    else:
        sys.stderr.write("Error: Please specify one of: build, query or simple modes. Use -h or --help for more information.\n")
        sys.exit(1)

    sys.exit(0)

if __name__ == "__main__":
    main()



# Copyright 2019-2020 John Lees, Sam Horsfield


'''Functions to run Bifrost'''

import os, sys
import subprocess
import re
import tempfile

def rtab_format(unitig_map, input_colour_pref, outfile):
    """Creates RTAB file.

    Args:
        unitig_map (dictionary)
            dictionary mapping each unitig to a binary colour vector.
        input_colour_pref (list)
            list of input colour file prefixs.
        outfile (str)
            output prefix for pyseer file
    """
    with open(outfile, "w") as o:
        o.write("Unitig_sequence")
        for index, colour_name in enumerate(input_colour_pref):
            o.write("\t" + colour_name)
            if index == len(input_colour_pref) - 1:
                o.write("\n")
        for unitig, colours_vector in unitig_map.items():
            o.write(unitig)
            for index, colour in enumerate(colours_vector):
                colour = int(colour == True)
                o.write("\t" + str(colour))
                if index == len(colours_vector) - 1:
                    o.write("\n")

def pyseer_format(unitig_map, input_colour_pref, outfile):
    """Creates pyseer file.

    Args:
        unitig_map (dictionary)
            dictionary mapping each unitig to a binary colour vector.
        input_colour_pref (list)
            list of input colour file prefixs.
        outfile (str)
            output prefix for pyseer file
    """
    with open(outfile, "w") as o:
        for unitig, colours_vector in unitig_map.items():
            o.write(unitig + " " + "|")
            for index, colour in enumerate(colours_vector):
                if colour:
                    o.write(" " + input_colour_pref[index] + ":1")
                if index == len(colours_vector) - 1:
                    o.write("\n")



from codecs import open
from os import path
import os
import re
import sys
import io
import platform
import subprocess

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion

def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)

class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(re.search(r'version\s*([\d.]+)', out.decode()).group(1))
            if cmake_version < '3.1.0':
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable,
                      '-DCMAKE_VERBOSE_MAKEFILE:BOOL=ON']

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j1']

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              str(find_version("unitig_caller/__init__.py")))
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='unitig-caller',
    version=find_version("unitig_caller/__init__.py"),
    description='unitig-caller: wrapper around mantis to detect presence of sequence elements',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/johnlees/unitig-caller',
    author='John Lees',
    author_email='john@johnlees.me',
    license='Apache Software License',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    keywords='gwas bacteria k-mer unitig',
    packages=['unitig_caller'],
    entry_points={
        "console_scripts": [
            'unitig-caller = unitig_caller.__main__:main'
            ]
    },
    ext_modules=[CMakeExtension('unitig_query')],
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False
)



#!/usr/bin/env python

import os, sys
import argparse

description = 'Test unitig calls match'
parser = argparse.ArgumentParser(description=description,
                                    prog='test-calls')

parser.add_argument('--method', help="File type", required=True)
parser.add_argument('--test', help="Calls produced", required=True)
parser.add_argument('--expected', help="Calls expected", required=True)

options = parser.parse_args()

if (options.method == "rtab"):
    with open(options.expected, 'r') as expected_rtab, open(options.test, 'r') as test_rtab:
        expected_samples = expected_rtab.readline().rstrip().split("\t")[1:]
        test_samples = test_rtab.readline().rstrip().split("\t")[1:]
        if (len(expected_samples) != len(test_samples)):
            sys.stderr.write("Length of rtab header mismatches\n")
            raise RuntimeError("Length of rtab header mismatches")
        elif len(set(expected_samples).intersection(test_samples)) != len(expected_samples):
            sys.stderr.write("Different samples contained\n")
            raise RuntimeError("Different samples contained")

        for expected_line, test_line in zip(expected_rtab, test_rtab):
            expected_calls = expected_line.rstrip().split("\t")[1:]  
            test_calls = test_line.rstrip().split("\t")[1:]
            for sample_idx, call in enumerate(expected_calls):
                if call != test_calls[test_samples.index(expected_samples[sample_idx])]:
                    sys.stderr.write("Calls mismatch for " + expected_samples[sample_idx] + "\n")
                    raise RuntimeError("Calls mismatch for " + expected_samples[sample_idx])

elif (options.method == "pyseer"):
    d = {}
    with open(options.expected, 'r') as expected_pyseer:
        for call in expected_pyseer:
            call_fields = call.rstrip().split(" ")
            d[call_fields[0]] = set()
            for sample in call_fields[2:]:
                d[call_fields[0]].add(sample)

    with open(options.test, 'r') as test_pyseer:
        for call in test_pyseer:
            call_fields = call.rstrip().split(" ")
            if not call_fields[0] in d.keys():
                sys.stderr.write("No call for " + call_fields[0] + "\n")
                raise RuntimeError("No call for " + call_fields[0])
            test_calls = set()
            for sample in call_fields[2:]:
                test_calls.add(sample)

            if test_calls != d[call_fields[0]]:
                sys.stderr.write("Calls mismatch for " + call_fields[0] + "\n")
                raise RuntimeError("Calls mismatch for " + call_fields[0])

else:
    raise RuntimeError("Unknown method type")

